###############################################################################
## This file is part of the Wolfpack module system
## Copyright (C) 1998-2001  Tothwolf <tothwolf@concentric.net>
## and is distributed according to the GNU general public license.
## For full details, read the top of wolfpack.tcl or the file called
## COPYING that was distributed with this code.
##
## $Id: auth.tcl,v 1.7 2001/06/06 09:36:14 tothwolf Exp $
##

# name: auth
# version: 1.4.0.0
# config: auth.conf
# description: Provides a login/logout authentication system and functions \
#	other modules can use for secure /msg and public commands.
# requires: isbotnick botonchan wpLog commandChar putList putnot \
#	convertSeconds formatTime splitList isNumber clearFields wipeFields \
#	moveFields whoisFields validHandle floodCheck
# provides: authUser authCheck

namespace eval auth {

  ##
  ## Login for a user: msg
  ##
  ## Args:    nick, user@host, handle, arg
  ## Returns: 0
  ##
  proc msg:login {nick uhost hand arg} {
    variable NamespaceCurrent
    variable configData

    if {([floodCheck $nick $uhost login]) && \
        ([validHandle $hand]) && (![isbotnick $nick])} then {
      set arg [split [string trimright $arg]]
      set pass [lindex $arg 0]
      set chan [lindex $arg 1]
      if {![string compare "" $pass]} then {
        putnot $nick "Usage: [commandChar msg]login <password> \[channel\]"
      } elseif {[passwdok $hand ""]} then {
        putnot $nick "You must set a password before you can use the 'login' command."
      } elseif {[string compare "" $chan]} then {
        if {[passwdok $hand $pass]} then {
          if {[validchan $chan]} then {
            if {[botonchan $chan]} then {
              if {[onchan $nick $chan]} then {
                if {![matchattr $hand |$configData(flag) $chan]} then {
                  authUser add $nick $uhost $hand $chan
                  putnot $nick "Login successful for $chan."
                } else {
                  putnot $nick "You are already logged in for $chan."
                }
              } else {
                putnot $nick "You are not on $chan."
              }
            } else {
              putnot $nick "I'm not on $chan."
            }
          } else {
            putnot $nick "I don't monitor $chan."
          }
        } else {
          wpLog o * $NamespaceCurrent "Invalid login from: $nick!$uhost ($hand) for $chan"
          putnot $nick "Login failed for $chan."
        }
      } else {
        set botonchan 0
        set onchan 0
        if {[passwdok $hand $pass]} then {
          foreach chan [channels] {
            if {[botonchan $chan]} then {
              set botonchan 1
              if {[onchan $nick $chan]} then {
                set onchan 1
                if {[authUser check $nick $uhost $hand $chan] != 1} then {
                  authUser add $nick $uhost $hand $chan
                  lappend list1 $chan,
                } else {
                  lappend list2 $chan,
                }
              }
            }
          }
          if {$botonchan} then {
            if {$onchan} then {
              if {[info exists list1]} then {
                putList putnot $nick [splitList $list1 160 "," ", "] "Login successful for: " "(cont): "
              } elseif {[info exists list2]} then {
                putList putnot $nick [splitList $list2 160 "," ", "] "You are already logged in for: " "(cont): "
              }
            } else {
              putnot $nick "You are not on any channels that I monitor."
            }
          } else {
            putnot $nick "I'm not on any channels."
          }
        } else {
          wpLog o * $NamespaceCurrent "Invalid login from: $nick!$uhost ($hand)"
          putnot $nick "Login failed."
        }
      }
      putloglev c * "($nick!$uhost) !$hand! login"
    }
    return 0
  }

  ##
  ## Logout for a user: msg
  ##
  ## Args:    nick, user@host, handle, arg
  ## Returns: 0
  ##
  proc msg:logout {nick uhost hand arg} {
    variable NamespaceCurrent
    variable configData

    if {([floodCheck $nick $uhost logout]) && \
        ([validHandle $hand]) && (![isbotnick $nick])} then {
      set arg [split [string trimright $arg]]
      set pass [lindex $arg 0]
      set chan [lindex $arg 1]
      if {![string compare "" $pass]} then {
        putnot $nick "Usage: [commandChar msg]logout <password> \[channel\]"
      } elseif {[passwdok $hand ""]} then {
        putnot $nick "You must set a password before you can use the 'logout' command."
      } elseif {[string compare "" $chan]} then {
        if {[passwdok $hand $pass]} then {
          if {[validchan $chan]} then {
            if {[botonchan $chan]} then {
              if {[onchan $nick $chan]} then {
                if {[matchattr $hand |$configData(flag) $chan]} then {
                  authUser del $nick $uhost $hand $chan
                  putnot $nick "Logout successful for $chan."
                } else {
                  putnot $nick "You were never logged in for $chan."
                }
              } else {
                putnot $nick "You are not on $chan."
              }
            } else {
              putnot $nick "I'm not on $chan."
            }
          } else {
            putnot $nick "I don't monitor $chan."
          }
        } else {
          wpLog o * $NamespaceCurrent "Invalid password for logout from: $nick!$uhost ($hand) for $chan"
          putnot $nick "Logout failed for $chan."
        }
      } else {
        set botonchan 0
        set onchan 0
        if {[passwdok $hand $pass]} then {
          foreach chan [channels] {
            if {[botonchan $chan]} then {
              set botonchan 1
              if {[onchan $nick $chan]} then {
                set onchan 1
                if {[matchattr $hand |$configData(flag) $chan]} then {
                  authUser del $nick $uhost $hand $chan
                  lappend list $chan,
                }
              }
            }
          }
          if {$botonchan} then {
            if {$onchan} then {
              if {[info exists list]} then {
                putList putnot $nick [splitList $list 160 "," ", "] "Logout successful for: " "(cont): "
              } else {
                putnot $nick "You were not logged in on any channels."
              }
            } else {
              putnot $nick "You are not on any channels that I monitor."
            }
          } else {
            putnot $nick "I'm not on any channels."
          }
        } else {
          wpLog o * $NamespaceCurrent "Invalid login from: $nick!$uhost ($hand)"
          putnot $nick "Login failed."
        }
      }
      putloglev c * "($nick!$uhost) !$hand! logout"
    }
    return 0
  }

  ##
  ## Login for a user: dcc
  ##
  ## Args:    handle, idx, arg
  ## Returns: 1
  ##
  proc dcc:login {hand idx arg} {
    variable configData

    set chan [lindex [split [string trimright $arg]] 0]
    if {[string compare "" $chan]} then {
      if {[validchan $chan]} then {
        if {[botonchan $chan]} then {
          if {[handonchan $hand $chan]} then {
            set nick [hand2nick $hand $chan]
            set uhost [getchanhost $nick $chan]
            if {![matchattr $hand |$configData(flag) $chan]} then {
              authUser add $nick $uhost $hand $chan
              putdcc $idx "Login successful for $chan."
            } else {
              putdcc $idx "You are already logged in for $chan."
            }
          } else {
            putdcc $idx "You are not on $chan."
          }
        } else {
          putdcc $idx "I'm not on $chan."
        }
      } else {
        putdcc $idx "I don't monitor $chan."
      }
    } else {
      set botonchan 0
      set handonchan 0
      foreach chan [channels] {
        if {[botonchan $chan]} then {
          set botonchan 1
          if {[handonchan $hand $chan]} then {
            set handonchan 1
            set nick [hand2nick $hand $chan]
            set uhost [getchanhost $nick $chan]
            if {[authUser check $nick $uhost $hand $chan] != 1} then {
              authUser add $nick $uhost $hand $chan
              lappend list1 $chan,
            } else {
              lappend list2 $chan,
            }
          }
        }
      }
      if {$botonchan} then {
        if {$handonchan} then {
          if {[info exists list1]} then {
            putList putdcc $idx [splitList $list1 160 "," ", "] "Login successful for: " "(cont): "
          } elseif {[info exists list2]} then {
            putList putdcc $idx [splitList $list2 160 "," ", "] "You are already logged in for: " "(cont): "
          }
        } else {
          putdcc $idx "You are not on any channels that I monitor."
        }
      } else {
        putdcc $idx "I'm not on any channels."
      }
    }
    return 1
  }

  ##
  ## Logout for a user: dcc
  ##
  ## Args:    handle, idx, arg
  ## Returns: 1
  ##
  proc dcc:logout {hand idx arg} {
    variable configData

    set chan [lindex [split [string trimright $arg]] 0]
    if {[string compare "" $chan]} then {
      if {[validchan $chan]} then {
        if {[botonchan $chan]} then {
          if {[handonchan $hand $chan]} then {
            set nick [hand2nick $hand $chan]
            set uhost [getchanhost $nick $chan]
            if {[matchattr $hand |$configData(flag) $chan]} then {
              authUser del $nick $uhost $hand $chan
              putdcc $idx "Logout successful for $chan."
            } else {
              putdcc $idx "You were never logged in for $chan."
            }
          } else {
            putdcc $idx "You are not on $chan."
          }
        } else {
          putdcc $idx "I'm not on $chan."
        }
      } else {
        putdcc $idx "I don't monitor $chan."
      }
    } else {
      set botonchan 0
      set handonchan 0
      foreach chan [channels] {
        if {[botonchan $chan]} then {
          set botonchan 1
          if {[handonchan $hand $chan]} then {
            set handonchan 1
            set nick [hand2nick $hand $chan]
            set uhost [getchanhost $nick $chan]
            if {[matchattr $hand |$configData(flag) $chan]} then {
              authUser del $nick $uhost $hand $chan
              lappend list $chan,
            }
          }
        }
      }
      if {$botonchan} then {
        if {$handonchan} then {
          if {[info exists list]} then {
            putList putdcc $idx [splitList $list 160 "," ", "] "Logout successful for: " "(cont): "
          } else {
            putdcc $idx "You are not logged in on any channels."
          }
        } else {
          putdcc $idx "You are not on any channels that I monitor."
        }
      } else {
        putdcc $idx "I'm not on any channels."
      }
    }
    return 1
  }

  ##
  ## Remove auth if user parts a channel
  ##
  ## Args:    nick, user@host, handle, channel, part message
  ## Returns: 0
  ##
  proc part:auth {nick uhost hand chan {msg ""}} {
    if {([validHandle $hand]) && (![isbotnick $nick]) && \
        ([authUser check $nick $uhost $hand $chan])} then {
      authUser del $nick $uhost $hand $chan
    }
    return 0
  }

  ##
  ## Remove auth if user signs off
  ##
  ## Args:    nick, user@host, handle, channel, arg
  ## Returns: 0
  ##
  proc sign:auth {nick uhost hand chan arg} {
    if {([validHandle $hand]) && (![isbotnick $nick]) && \
        ([authUser check $nick $uhost $hand $chan])} then {
      authUser del $nick $uhost $hand ""
    }
    return 0
  }

  ##
  ## Remove auth if user signs off due to a netsplit
  ##
  ## Args:    nick, user@host, handle, channel
  ## Returns: 0
  ##
  proc splt:auth {nick uhost hand chan} {
    if {([validHandle $hand]) && (![isbotnick $nick]) && \
        ([authUser check $nick $uhost $hand $chan])} then {
      authUser del $nick $uhost $hand ""
    }
    return 0
  }

  ##
  ## Change auth when a user changes nicks: delete
  ##
  ## Args:    nick, user@host, handle, channel, newnick
  ## Returns: 0
  ##
  proc nick:auth {nick uhost hand chan newnick} {
    if {([validHandle $hand]) && (![isbotnick $nick])} then {
      if {[set check [authUser check $nick $uhost $hand $chan]] == 1} then {
        authUser update $nick $uhost $hand $chan $newnick
      } elseif {$check == -1} then {
        authUser del $nick $uhost $hand $chan
      }
    }
    return 0
  }

  ##
  ## Remove expired logins
  ##
  ## Args:    minute, hour, day, month, year
  ## Returns: 0
  ##
  proc time:10Minutes {minute hour day month year} {
    variable configData

    if {[string compare "" [set users [userlist $configData(flag)]]]} then {
      foreach user $users {
        if {($configData(expire)) && \
            ([isNumber [set LoginTime [getuser $user XTRA $configData(logintime)]]]) && \
            ([expr [clock seconds] - $LoginTime] >= [convertSeconds from minutes $configData(expire)])} then {
          authUser del "" "" $user ""
        }
      }
    }
    return 0
  }

  ##
  ## Handle auth check, add, and del for a given user and channel
  ##
  ## Args:    action {check,add,del}, nick, user@host, handle, channel
  ## Returns: 1 if action successful,
  ##          0 otherwise
  ##
  proc authUser {action nick uhost hand chan args} {
    variable configData

    switch -exact -- $action {
      check {
        if {[string compare "" $chan]} then {
          if {([matchattr $hand |$configData(flag) $chan]) && \
              (![string compare "$nick!$uhost" \
                 [getuser $hand XTRA $configData(authfield)]])} then {
            set LoginTime [getuser $hand XTRA $configData(logintime)]
            if {(!$configData(expire)) || \
                (([isNumber $LoginTime]) && \
                 ([expr [clock seconds] - $LoginTime] < [convertSeconds from minutes $configData(expire)]))} then {
              return 1
            } else {
              return -1
            }
          }
        } elseif {([matchattr $hand $configData(flag)]) && \
                  (![string compare "$nick!$uhost" \
                     [getuser $hand XTRA $configData(authfield)]])} then {
          set LoginTime [getuser $hand XTRA $configData(logintime)]
          if {(!$configData(expire)) || \
              (([isNumber $LoginTime]) && \
               ([expr [clock seconds] - $LoginTime] < [convertSeconds from minutes $configData(expire)]))} then {
            return 1
          } else {
            return -1
          }
        }
      }
      update {
        if {(![matchattr $hand $configData(flag)]) || \
            (![matchattr $hand |$configData(flag) $chan])} then {
          chattr $hand +$configData(flag)|+$configData(flag) $chan
        }
        if {[string compare "" [set newnick [lindex $args 0]]]} then {
          if {[string compare $newnick!$uhost \
               [getuser $hand XTRA $configData(authfield)]]} then {
            setuser $hand XTRA [string toupper $configData(authfield)] $newnick!$uhost
          }
        }
      }
      add {
        if {(![matchattr $hand $configData(flag)]) || \
            (![matchattr $hand |$configData(flag) $chan])} then {
          chattr $hand +$configData(flag)|+$configData(flag) $chan
        }
        if {[string compare $nick!$uhost \
             [getuser $hand XTRA $configData(authfield)]]} then {
          setuser $hand XTRA [string toupper $configData(authfield)] $nick!$uhost
        }
        set time [clock seconds]
        setuser $hand XTRA [string toupper $configData(logintime)] $time
        setuser $hand XTRA [string toupper $configData(loginfield)] [formatTime $time]
        if {$configData(expire)} then {
          setuser $hand XTRA [string toupper $configData(expirefield)] \
           [formatTime [expr $time + [convertSeconds from minutes $configData(expire)]]]
        } else {
          clearFields $hand $configData(expirefield)
        }
        return 1
      }
      del {
        if {[string compare "" $chan]} then {
          if {[matchattr $hand |$configData(flag) $chan]} then {
            chattr $hand |-$configData(flag) $chan
          }
          set ok 1
          foreach chan [channels] {
            if {[matchattr $hand |$configData(flag) $chan]} then {
              if {![handonchan $hand $chan]} then {
                chattr $hand |-$configData(flag) $chan
              } else {
                set ok 0
              }
            }
          }
          if {$ok} then {
            if {[matchattr $hand $configData(flag)]} then {
              chattr $hand -$configData(flag)
            }
            clearFields $hand "$configData(authfield) $configData(loginfield) $configData(expirefield) $configData(logintime)"
          }
        } else {
          foreach chan [channels] {
            if {[matchattr $hand |$configData(flag) $chan]} then {
              chattr $hand |-$configData(flag) $chan
            }
          }
          if {[matchattr $hand $configData(flag)]} then {
            chattr $hand -$configData(flag)
          }
          clearFields $hand "$configData(authfield) $configData(loginfield) $configData(expirefield) $configData(logintime)"
        }
        return 1
      }
    }
    return 0
  }

  ##
  ## Check to see if a user is logged in, and notice them if not
  ##
  ## Args:    type {pub,msg,dcc,fil}, nick, user@host, handle, channel,
  ##           command
  ## Returns: 1 if user is logged in, or mode is dcc or fil
  ##          0 otherwise
  ##
  proc authCheck {type command nick uhost hand chan} {
    variable NamespaceCurrent

    switch -exact -- $type {
      pub -
      msg {
        set check [authUser check $nick $uhost $hand $chan]
        if {$check == 1} then {
          return 1
        } else {
          switch -exact -- $type {
            pub {
              set Log "<<$nick>> !$hand!"
            }
            msg {
              set Log "($nick!$uhost) !$hand!"
            }
          }
          if {$check == -1} then {
            wpLog c * $NamespaceCurrent "$Log Failed $command (login expired)"
            putnot $nick "Your login has expired, please login again."
          } else {
            wpLog c * $NamespaceCurrent "$Log Failed $command (not logged in)"
            putnot $nick "You must login to use the command: [commandChar $type]$command"
          }
          putnot $nick "To login, type: [commandChar msg]login <password> \[channel\]"
        }
      }
      dcc {
        return 1
      }
      fil {
        return 1
      }
      default {
        error "Unknown type: $type"
      }
    }
    return 0
  }

  ##
  ## Trace configuration changes
  ##
  ## Args:    name1, name2, operation
  ## Returns: nothing
  ##
  proc TraceConfigData {name1 name2 op} {
    variable configData
    variable oldConfigData

    if {[string compare $oldConfigData($name2) $configData($name2)]} then {
      switch -exact -- $name2 {
        flag {
          foreach user [userlist $oldConfigData(flag)] {
            chattr $user -[set oldConfigData(flag)]+[set configData(flag)]
          }
          foreach chan [channels] {
            foreach user [userlist |$oldConfigData(flag) $chan] {
              chattr $user |-[set oldConfigData(flag)]+[set configData(flag)] $chan
            }
          }
        }
        expire {
          if {([isNumber $configData(expire)]) && ($configData(expire))} then {
            foreach user [userlist] {
              if {[string compare "" [set LoginTime [getuser $user XTRA $configData(logintime)]]]} then {
                setuser $user XTRA [string toupper $configData(expirefield)] \
                 [formatTime [expr $LoginTime + [convertSeconds from minutes $configData(expire)]]]
              }
            }
          } else {
            wipeFields $configData(expirefield)
          }
        }
        authfield -
        loginfield -
        expirefield {
          moveFields $oldConfigData($name2) [string toupper $configData($name2)]
          whoisFields update $configData($name2) $oldConfigData($name2)
        }
        logintime {
          moveFields $oldConfigData(logintime) [string toupper $configData(logintime)]
        }
      }
      set oldConfigData($name2) $configData($name2)
    }
    return
  }

  ##
  ## Trace and respond to variable changes
  ##
  ## Args:    action {variable|vdelete}
  ## Returns: nothing
  ##
  proc TraceTable {action} {
    variable NamespaceCurrent
    variable configData

    trace $action configData w ${NamespaceCurrent}::TraceConfigData
    return
  }

  ##
  ## Init module specific things at load time
  ##
  ## Args:    action {load|unload}
  ## Returns: nothing
  ##
  proc InitTable {action} {
    variable NamespaceCurrent
    variable configData
    variable oldConfigData

    switch -exact -- $action {
      load {
        if {![info exists NamespaceCurrent]} then {
          set NamespaceCurrent [namespace current]
        }
        foreach {name data} [array get configData] {
          if {![info exists oldConfigData($name)]} then {
            set oldConfigData($name) $data
          }
        }
        whoisFields update "$configData(authfield) $configData(loginfield) $configData(expirefield)" \
                           "$oldConfigData(authfield) $oldConfigData(loginfield) $oldConfigData(expirefield)"
      }
      unload {
        whoisFields remove "$configData(authfield) $configData(loginfield) $configData(expirefield)"
      }
    }
    return
  }

  ##
  ## Bind procs to commands and events
  ##
  ## Args:    action {bind|unbind}, what {type|*}, args {...}
  ## Returns: nothing
  ##
  proc BindTable {action what args} {
    variable NamespaceCurrent

    if {![string compare * $what]} then {
      set what "msg dcc sign part nick splt time"
    }
    foreach type $what {
      switch -exact -- $type {
        msg {
          $action msg -|- login ${NamespaceCurrent}::msg:login
          $action msg -|- logon ${NamespaceCurrent}::msg:login
          $action msg -|- logout ${NamespaceCurrent}::msg:logout
          $action msg -|- logoff ${NamespaceCurrent}::msg:logout
        }
        dcc {
          $action dcc -|- login ${NamespaceCurrent}::dcc:login
          $action dcc -|- logon ${NamespaceCurrent}::dcc:login
          $action dcc -|- logout ${NamespaceCurrent}::dcc:logout
          $action dcc -|- logoff ${NamespaceCurrent}::dcc:logout
        }
        sign {
          $action sign -|- * ${NamespaceCurrent}::sign:auth
        }
        part {
          $action part -|- * ${NamespaceCurrent}::part:auth
        }
        nick {
          $action nick -|- * ${NamespaceCurrent}::nick:auth
        }
        splt {
          $action splt -|- * ${NamespaceCurrent}::splt:auth
        }
        time {
          $action time -|- "?0 * * * *" ${NamespaceCurrent}::time:10Minutes
        }
      }
    }
    return
  }

  # Configuration data defaults
  array set configDataDefaults {
    flag L
    expire 240
    authfield authuser
    loginfield authlogin
    expirefield authexpire
    logintime authtime
  }

  # Configuration data descriptions
  array set configDataDesc {
    flag "\[flag\] flag auth code will add when a user is auth'd"
    expire "\[minutes\] minutes till login expires"
    authfield "\[field\] field used to verify a user's user@host"
    loginfield "\[field\] field used to display login time for a user"
    expirefield "\[field\] field used to display login expire time for a user"
    logintime "\[field\] field used to store login time for a user"
  }

} ;# namespace auth

return
