# botnetop.tcl v2.10 (16 June 2000)
# copyright (c) 1998-2000 by slennox <slennox@egghelp.org>
# slennox's eggdrop page - http://www.egghelp.org/
#
# Cleanup and conversion to a wolfpack module (04 July 2000)
#  by Tothwolf (tothwolf@concentric.net)
#
# name: botnetop
# version: 2.10.5.1
# config: botnetop.conf
# description: Allows bots to op each other via the botnet
# requires: botonchan isbotnick islinked wpLog
#
# $Id: botnetop.tcl,v 1.2 2001/06/07 08:19:26 tothwolf Exp $
#
# This script was written mainly as an alternative to getops.tcl. It allows
# linked bots to securely give ops to one another on their channels. It
# also automatically invites linked bots to channels which are invite-only,
# shares keys with linked bots for channels which are keyed, raises a
# channel's limit if a bot can't get in because the channel is full, unbans
# a bot if it can't get in because it matches a ban, and automatically adds
# new hostmasks when a bot's user@host info has changed. There is a delayed
# op mechanism to prevent mode +o flooding whenever a bot joins a channel,
# and op/invite checking routines are significantly more secure than those
# in getops.tcl.
#
# botnetop.tcl v2.xx is compatible with bots running older versions of
# botnetop.tcl (v1.00 and up) and bots running the botnetop.tcl component
# of netbots.tcl.
#
# Note that this script was developed on the eggdrop 1.3/1.4 series and may
# not work properly on other versions.
#
# v2.00 - New standalone release. Contains refinements and features from
#         the netbots.tcl version of botnetop.tcl.
# v2.02 - Improved support for leading +-^= characters in hostmasks.
#         Fixed 'islinked' problem on eggdrop 1.3.24 and earlier.
# v2.10 - Requests to unopped bots were causing high CPU usage and botnet
#         traffic on some large opless botnets - rewrote proc to request
#         from opped bots only.
#         Added bop_modeop feature to provide faster opping.
#         Convinced myself to change putquick back to pushmode (suits
#         bop_modeop better).
#
# About the delay and maxreq settings: botnetop.tcl uses two
# methods to reduce/eliminate mode +o flooding - delay is used by the
# bot giving ops, while maxreq is used by the bot asking for ops. You
# can use both features, one or the other, or disable both. Below is an
# explanation of some possible combinations.
#
# delay 20; maxreq 0: all bots will ask a non-opped bot if it needs
# ops after a delay of between 0-20 seconds (different bots will delay for
# different amounts of time). The bot that needs ops will respond to all
# offers.
#
# delay 0; maxreq 2: all bots will ask a non-opped bot if it needs
# ops without any delay. The bot that needs ops will respond to a maximum
# of 2 offers.
#
# delay 20; maxreq 3: all bots will ask a non-opped bot if it needs
# ops after a delay of between 0-20 seconds. The bot that needs ops will
# respond to a maximum of 3 offers.
#
# delay 0; maxreq 0: all bots will ask a non-opped bot if it needs
# ops without any delay. The bot that needs ops will respond to all offers.

namespace eval botnetop {

  ##
  ## Called by the bot's need-* options when the bot needs in a channel
  ##
  ## Args:    what, channel
  ## Returns: nothing
  ##
  proc letmein {what chan} {
    global botname botnick
    variable NamespaceCurrent
    variable configData
    variable NeedKey
    variable NeedInvite
    variable NeedLimit
    variable NeedUnban

    if {([string compare "" [bots]]) && ([botonchan $chan])} then {
      set chanLower [string tolower $chan]
      switch -exact -- $what {
        needkey {
          if {!$NeedKey($chanLower)} then {
            set list ""
            foreach bot [bots] {
              if {([matchattr $bot b]) && ([matchattr $bot o|o $chan])} then {
                putbot $bot "wantkey $chan $botnick"
                lappend list $bot
              }
            }
            if {($configData(log) >= 1) && ([string compare "" $list])} then {
              regsub -all -- " " $list ", " list
              wpLog o * $NamespaceCurrent "requested key for $chan from $list"
            }
            set NeedKey($chanLower) 1
            utimer 30 [split "set ${NamespaceCurrent}::NeedKey($chanLower) 0"]
          }
        }
        needinvite {
          if {!$NeedInvite($chanLower)} then {
            set list ""
            foreach bot [bots] {
              if {([matchattr $bot b]) && ([matchattr $bot o|o $chan])} then {
                if {[string compare "" $botname]} then {
                  putbot $bot "wantinvite $chan $botnick [string trimleft [lindex [split $botname !] 1] "~+-^="]"
                } else {
                  putbot $bot "wantinvite $chan $botnick"
                }
                lappend list $bot
              }
            }
            if {($configData(log) >= 1) && ([string compare "" $list])} then {
              regsub -all -- " " $list ", " list
              wpLog o * $NamespaceCurrent "requested invite to $chan from $list"
            }
            set NeedInvite($chanLower) 1
            utimer 30 [split "set ${NamespaceCurrent}::NeedInvite($chanLower) 0"]
          }
        }
        needlimit {
          if {!$NeedLimit($chanLower)} then {
            set list ""
            foreach bot [bots] {
              if {([matchattr $bot b]) && ([matchattr $bot o|o $chan])} then {
                putbot $bot "wantlimit $chan $botnick"
                lappend list $bot
              }
            }
            if {($configData(log) >= 1) && ([string compare "" $list])} then {
              regsub -all -- " " $list ", " list
              wpLog o * $NamespaceCurrent "requested limit raise on $chan from $list"
            }
            set NeedLimit($chanLower) 1
            utimer 30 [split "set ${NamespaceCurrent}::NeedLimit($chanLower) 0"]
          }
        }
        needunban {
          if {!$NeedUnban($chanLower)} then {
            set list ""
            foreach bot [bots] {
              if {([matchattr $bot b]) && ([matchattr $bot o|o $chan])} then {
                putbot $bot "wantunban $chan $botnick $botname"
                lappend list $bot
              }
            }
            if {($configData(log) >= 1) && ([string compare "" $list])} then {
              regsub -all -- " " [join $list] ", " list
              wpLog o * $NamespaceCurrent "requested unban on $chan from $list"
            }
            set NeedUnban($chanLower) 1
            utimer 30 [split "set ${NamespaceCurrent}::NeedUnban($chanLower) 0"]
          }
        }
      }
    }
    return
  }

  ##
  ## Request ops from a bot that gets op'd
  ##
  ## Args:    nick, user@host, handle, channel, mode, target
  ## Returns: 0
  ##
  proc mode:requestops {nick uhost hand chan mode target} {
    if {(![string compare "+o" $mode]) && (![botisop $chan]) && \
        (![isbotnick $nick]) && (![isbotnick $target]) && \
        ([matchattr [set targetHand [nick2hand $target $chan]] b]) && \
        ([matchattr $targetHand o|o $chan]) && \
        ([islinked $targetHand])} then {
      putbot $targetHand "reqops $chan"
    }
    return 0
  }

  ##
  ## Request ops from bots when this bot links to the botnet
  ##
  ## Args:    bot, via
  ## Returns: 0
  ##
  proc link:requestops {bot via} {
    global botnet-nick
    variable NamespaceCurrent

    if {(![string compare $bot ${botnet-nick}]) || \
        (![string compare $via ${botnet-nick}]) && \
        (![regexp .*RequestOpsChannels.* [utimers]])} then {
      utimer 2 ${NamespaceCurrent}::RequestOpsChannels
    }
    return 0
  }

  ##
  ## Answer a request if this bot wants ops
  ##
  ## Args:    bot name, command, arg
  ## Returns: 0
  ##
  proc bot:doyawantops {bot cmd arg} {
    global botname botnick
    variable NamespaceCurrent
    variable configData
    variable OpRequests

    set arg [split $arg]
    set chan [lindex $arg 0]
    set fromnick [lindex $arg 1]
    if {([validchan $chan]) && ([botonchan $chan]) && \
        (![botisop $chan])} then {
      set chanLower [string tolower $chan]
      if {($configData(maxreq)) && \
          ($OpRequests($chanLower) < $configData(maxreq)) && \
          ([onchan $fromnick $chan]) && (![onchansplit $fromnick $chan]) && \
          ([isop $fromnick $chan]) && \
          ([matchattr [nick2hand $fromnick $chan] o|o $chan]) && \
          ([islinked $bot])} then {
        putbot $bot "yesiwantops $chan $botnick [string trimleft [lindex [split $botname !] 1] "~+-^="]"
        incr OpRequests($chanLower)
        if {$configData(maxreq)} then {
          KillUtimerCommand "${NamespaceCurrent}::ResetOpRequests $chanLower"
          utimer 30 [split "${NamespaceCurrent}::ResetOpRequests $chanLower"]
        }
        if {$configData(log) >= 2} then {
          wpLog o * $NamespaceCurrent "requested ops from $bot on $chan"
        }
      }
    }
    return 0
  }

  ##
  ## Answer a request to a request if this bot wants ops
  ##
  ## Args:    bot name, command, arg
  ## Returns: 0
  ##
  proc bot:yesiwantops {bot cmd arg} {
    global strict-host
    variable NamespaceCurrent
    variable configData

    set arg [split $arg]
    set chan [lindex $arg 0]
    set fromnick [lindex $arg 1]
    set fromhost [lindex $arg 2]
    if {([botonchan $chan]) && ([botisop $chan]) && \
        ([onchan $fromnick $chan]) && (![onchansplit $fromnick $chan]) && \
        ([matchattr $bot b]) && ([matchattr $bot o|o $chan]) && \
        (![matchattr $bot d|d $chan])} then {
      if {(![matchattr [nick2hand $fromnick $chan] o|o $chan]) && \
          ([string compare "" $fromhost]) && \
          ([string compare $fromhost \
            [string trimleft [getchanhost $fromnick $chan] "~+-^="]]) && \
          ($configData(addhost))} then {
        if {!${strict-host}} then {
          set host *![string trimleft [getchanhost $fromnick $chan] "~+-^="]
        } else {
          set host *![getchanhost $fromnick $chan]
        }
        setuser $bot HOSTS $host
        wpLog o * $NamespaceCurrent "added host $host to $bot"
      }
      if {(($configData(osync)) || (![isop $fromnick $chan]))} then {
        pushmode $chan +o $fromnick
        if {$configData(log) >= 2} then {
          if {[string compare $fromnick $bot]} then {
            wpLog o * $NamespaceCurrent "gave ops to $bot (using nick $fromnick) on $chan"
          } else {
            wpLog o * $NamespaceCurrent "gave ops to $bot on $chan"
          }
        }
      }
    }
    return 0
  }

  ##
  ## Botnet reqops
  ##
  ## Args:    bot name, command, arg
  ## Returns: 0
  ##
  proc bot:reqops {bot cmd arg} {
    variable NamespaceCurrent
    variable configData
    variable AskTimer

    set chan [lindex [split $arg] 0]
    if {([validchan $chan]) && ([matchattr $bot b]) && \
        ([matchattr $bot o|o $chan]) && (![matchattr $bot d|d $chan])} then {
      set chanLower [string tolower $chan]
      if {![info exists AskTimer($bot:$chanLower)]} then {
        set AskTimer($bot:$chanLower) 1
        if {(!$configData(delay)) || ([LowBots $chan])} then {
          utimer 2 [split "${NamespaceCurrent}::AskBot $bot $chan"]
        } else {
          utimer [expr [rand $configData(delay)] + 2] [split "${NamespaceCurrent}::AskBot $bot $chan"]
        }
      }
    }
    return 0
  }

  ##
  ## Botnet wantkey/invite/limit/unban
  ##
  ## Args:    bot name, command, arg
  ## Returns: 0
  ##
  proc bot:want {bot cmd arg} {
    variable NamespaceCurrent
    variable configData
    variable WhoNicks

    set arg [split $arg]
    set chan [lindex $arg 0]
    if {([validchan $chan]) && ([botisop $chan]) && \
        ([matchattr $bot b]) && \
        ([matchattr $bot fo|fo $chan])} then {
      set fromhost [lindex $arg 2]
      switch -exact -- [split $cmd] {
        wantkey {
          if {[regexp .*k.* [lindex [split [getchanmode $chan]] 0]]} then {
            putbot $bot "thekey $chan [split [lindex [getchanmode $chan]] 1]"
            if {$configData(log) >= 1} then {
              wpLog o * $NamespaceCurrent "gave key for $chan to $bot"
            }
          }
        }
        wantinvite {
          set fromnick [lindex $arg 1]
          if {(($configData(icheck)) && ([string compare "" $fromhost]))} then {
            if {![info exists WhoNicks($fromnick)]} then {
              set WhoNicks($fromnick) "$chan $bot $fromhost"
              utimer 60 [split "${NamespaceCurrent}::WhoUnset $fromnick"]
            }
            putserv "WHO $fromnick"
          } else {
            putserv "INVITE $fromnick $chan"
            if {$configData(log) >= 1} then {
              if {[string compare $fromnick $bot]} then {
                wpLog o * $NamespaceCurrent "invited $bot (using nick $fromnick) to $chan"
              } else {
                wpLog o * $NamespaceCurrent "invited $bot to $chan"
              }
            }
          }
        }
        wantlimit {
          pushmode $chan +l [expr [llength [chanlist $chan]] + 1]
          if {$configData(log) >= 1} then {
            wpLog o * $NamespaceCurrent "raised limit on $chan as requested by $bot"
          }
        }
        wantunban {
          foreach {ban from time} [chanbans $chan] {
            if {[string match [string tolower $ban] \
                 [string tolower $fromhost]]} then {
              pushmode $chan -b $ban
              if {$configData(log) >= 1} then {
                wpLog o * $NamespaceCurrent "unbanned $bot on $chan"
              }
            }
          }
        }
      }
    }
    return 0
  }

  ##
  ## Botnet thekey
  ##
  ## Args:    bot name, command, arg
  ## Returns: 0
  ##
  proc bot:thekey {bot cmd arg} {
    variable NamespaceCurrent
    variable KeyJoin

    set arg [split $arg]
    set chan [lindex $arg 0]
    set chanLower [string tolower $chan]
    if {((![botonchan $chan]) && (!$KeyJoin($chanLower)))} then {
      putserv "JOIN $chan [lindex $arg 1]"
      set KeyJoin($chanLower) 1
      utimer 10 [split "set ${NamespaceCurrent}::KeyJoin($chanLower) 0"]
    }
    return 0
  }

  ##
  ## Ask a bot on join if they need ops, or set defaults
  ##
  ## Args:    nick, user@host, handle, channel
  ## Returns: 0
  ##
  proc join:askops {nick uhost hand chan} {
    variable NamespaceCurrent
    variable configData
    variable AskTimer

    if {![isbotnick $nick]} then {
      if {(([matchattr $hand b]) && ([matchattr $hand o|o $chan]) && \
           (![matchattr $hand d|d $chan]))} then {
        set chanLower [string tolower $chan]
        if {![info exists AskTimer($hand:$chanLower)]} then {
          set AskTimer($hand:$chanLower) 1
          if {((!$configData(delay)) || ([LowBots $chan]))} then {
            utimer 5 [split "${NamespaceCurrent}::AskBot $hand $chan"]
          } else {
            utimer [expr [rand $configData(delay)] + 5] [split "${NamespaceCurrent}::AskBot $hand $chan"]
          }
        }
      }
    } else {
      SetNeed $chan
    }
    return 0
  }

  ##
  ## raw 352 (who reply)
  ##
  ## Args:    from, keyword, arg
  ## Returns: 0
  ##
  proc raw:352 {from key arg} {
    variable NamespaceCurrent
    variable configData
    variable WhoNicks

    set arg [split $arg]
    set fromnick [lindex $arg 5]
    if {[info exists WhoNicks($fromnick)]} then {
      scan $WhoNicks($fromnick) "%s%s%s" chan bot fromhost
      unset WhoNicks($fromnick)
      if {![string compare $fromhost \
            [string trimleft \
             [lindex $arg 2]@[lindex $arg 3] "~+-^="]]} then {
        putserv "INVITE $fromnick $chan"
        if {$configData(log) >= 1} then {
          if {[string compare $fromnick $bot]} then {
            wpLog o * $NamespaceCurrent "invited $bot (using nick $fromnick) to $chan"
          } else {
            wpLog o * $NamespaceCurrent "invited $bot to $chan"
          }
        }
      }
    }
    return 0
  }

  ##
  ## Check for new channels every 5 minutes
  ##
  ## Args:    minute, hour, day,  month, year
  ## Returns: 0
  ##
  proc time:5Minutes {minute hour day month year} {
    variable NeedInvite

    foreach chan [channels] {
      if {![info exists NeedInvite([string tolower $chan])]} then {
        SetNeed $chan
      }
    }
    return 0
  }

  ##
  ## Kill all pending utimers for the given command
  ##
  ## Args:    command
  ## Returns: number of killed utimers
  ##
  proc KillUtimerCommand {cmd} {
    set cmd [split $cmd]
    foreach utimer [utimers] {
      foreach {seconds command id} $utimer {break}
      if {![string compare $cmd $command]} then {
        killutimer $id
      }
    }
    return
  }

  ##
  ## Set OpRequests for a given channel to 0
  ##
  ## Args:    channel
  ## Returns: nothing
  ##
  proc ResetOpRequests {chan} {
    variable OpRequests

    set OpRequests($chan) 0
    return
  }

  ##
  ## Check all channels to see if the bot needs ops
  ##
  ## Args:    none
  ## Returns: nothing
  ##
  proc RequestOpsChannels {} {
    foreach chan [channels] {
      if {([botonchan $chan]) && (![botisop $chan])} then {
        requestops $chan
      }
    }
    return
  }

  ##
  ## Request ops from op'd bots on the given channel
  ##
  ## Args:    channel
  ## Returns: nothing
  proc requestops {chan} {
    foreach bot [chanlist $chan b] {
      if {([isop $bot $chan]) && \
          ([matchattr [set hand [nick2hand $bot $chan]] o|o $chan]) && \
          ([islinked $hand])} then {
        putbot $hand "reqops $chan"
      }
    }
    return
  }

  ##
  ## If the given nick exists in WhoNicks, unset it
  ##
  ## Args:    nick
  ## Returns: nothing
  ##
  proc WhoUnset {nick} {
    variable WhoNicks

    if {[info exists WhoNicks($nick)]} then {
      unset WhoNicks($nick)
    }
    return
  }

  ##
  ## Count number of op'd bots in the given channel
  ##
  ## Args:    channel
  ## Returns: 1 if 3 or more bots are op'd in the given channel
  ##          0 otherwise
  ##
  proc LowBots {chan} {
    set bots 1
    foreach bot [chanlist $chan b] {
      if {(![isbotnick $bot]) && ([isop $bot $chan])} then {
        incr bots
      }
    }
    if {$bots < 3} then {
      return 1
    }
    return 0
  }

  ##
  ## AskBot
  ##
  ## Args:    handle, channel
  ## Returns: nothing
  ##
  proc AskBot {hand chan} {
    global botnick
    variable AskTimer

    set chanLower [string tolower $chan]
    if {[info exists AskTimer($hand:$chanLower)]} then {
      unset AskTimer($hand:$chanLower)
    }
    if {([islinked $hand]) && ([validchan $chan]) && ([botonchan $chan]) && \
        ([botisop $chan]) && ([matchattr $hand b]) && \
        ([matchattr $hand o|o $chan]) && (![matchattr $hand d|d $chan])} then {
      putbot $hand "doyawantops $chan $botnick"
    }
    return
  }

  ##
  ## Set temporary variables to default values and
  ## set need-* options for the given channel
  ##
  ## Args:    channel
  ## Returns: nothing
  ##
  proc SetNeed {chan} {
    variable NamespaceCurrent
    variable OpRequests
    variable NeedKey
    variable NeedInvite
    variable NeedLimit
    variable NeedUnban
    variable KeyJoin

    set chanLower [string tolower $chan]
    set OpRequests($chanLower) 0
    set NeedKey($chanLower) 0
    set NeedInvite($chanLower) 0
    set NeedLimit($chanLower) 0
    set NeedUnban($chanLower) 0
    set KeyJoin($chanLower) 0
    channel set $chan need-op [split "${NamespaceCurrent}::requestops $chan"]
    channel set $chan need-key [split "${NamespaceCurrent}::letmein needkey $chan"]
    channel set $chan need-invite [split "${NamespaceCurrent}::letmein needinvite $chan"]
    channel set $chan need-limit [split "${NamespaceCurrent}::letmein needlimit $chan"]
    channel set $chan need-unban [split "${NamespaceCurrent}::letmein needunban $chan"]
    return
  }

  ##
  ## Init module specific things at load time
  ##
  ## Args:    action {load|unload}
  ## Returns: nothing
  ##
  proc InitTable {action} {
    variable NamespaceCurrent

    switch -exact -- $action {
      load {
        if {![info exists NamespaceCurrent]} then {
          set NamespaceCurrent [namespace current]
        }
      }
    }
    return
  }

  ##
  ## Bind procs to commands and events
  ##
  ## Args:    action {bind|unbind}, what {type|*}, args {...}
  ## Returns: nothing
  ##
  proc BindTable {action what args} {
    variable NamespaceCurrent

    if {![string compare * $what]} then {
      set what "mode link bot join raw time"
    }
    foreach type $what {
      switch -exact -- $type {
        mode {
          $action mode -|- * ${NamespaceCurrent}::mode:requestops
        }
        link {
          $action link -|- * ${NamespaceCurrent}::link:requestops
        }
        bot {
          $action bot -|- doyawantops ${NamespaceCurrent}::bot:doyawantops
          $action bot -|- yesiwantops ${NamespaceCurrent}::bot:yesiwantops
          $action bot -|- reqops ${NamespaceCurrent}::bot:reqops
          $action bot -|- wantkey ${NamespaceCurrent}::bot:want
          $action bot -|- wantinvite ${NamespaceCurrent}::bot:want
          $action bot -|- wantlimit ${NamespaceCurrent}::bot:want
          $action bot -|- wantunban ${NamespaceCurrent}::bot:want
          $action bot -|- thekey ${NamespaceCurrent}::bot:thekey
        }
        join {
          $action join -|- * ${NamespaceCurrent}::join:askops
        }
        raw {
          $action raw -|- 352 ${NamespaceCurrent}::raw:352
        }
        time {
          $action time -|- "?0 * * * *" ${NamespaceCurrent}::time:5Minutes
          $action time -|- "?5 * * * *" ${NamespaceCurrent}::time:5Minutes
        }
      }
    }
    return
  }

  # Configuration data defaults
  array set configDataDefaults {
    delay 20
    maxreq 3
    modeop 0
    linkop 1
    icheck 0
    osync 0
    addhost 0
    log 2
  }

  # Configuration data descriptions
  array set configDataDesc {
    delay "Maximum number of seconds to delay before asking a non-opped bot if it needs ops. This is used by the bot giving ops. Higher settings help reduce +o flooding when a bot joins. Note that the delay is automatically skipped if there are fewer than 3 bots opped on the channel. Valid settings: 0 to disable delayed op, otherwise 1 or higher."
    maxreq "Maximum number of bots to request ops from at a time. This is used by the bot requesting ops. This feature can completely eliminate +o flooding when a bot joins, but the bot will request ops from fewer bots. Valid settings: 0 to disable, otherwise 1 or higher."
    modeop "Make the bot send an op request when it sees a bot become opped on a channel? This will make your bot get ops sooner when another bot is given ops. Valid settings: 1 to enable, 0 to disable."
    linkop "Make the bot send an op request when a bot links? This will make your bot get ops sooner if linking is delayed, and provides a better chance of the bots opping one another if there are linking problems. Valid settings: 1 to enable, 0 to disable."
    icheck "Check a bot's user@host before inviting it to a channel? Normally, the bot can sometimes invite the wrong user to +i channel if there are splits occurring. Enabling this check can prevent that. Valid settings: 1 to enable, 0 to disable."
    osync "Give ops to a bot even if it looks like it's already opped on the channel? Normally, the script will ignore a request for ops from a bot if that bot is already opped. But in some desync situations, a bot may appear to be opped even when it isn't, so you might want to enable this if you often have desync problems. Note that enabling this can result in +o flooding. Valid settings: 1 to enable, 0 to disable."
    addhost "Let the script automatically add new hostmasks for bots? (e.g. if you change a bot's vhost). This feature doesn't work all the time due to complications with protect-telnet, strict-host, bot host sharing flaws, security concerns, etc., but it's convenient when it works. Valid settings: 1 to enable, 0 to disable."
    log "By default, botnetop.tcl logs all ops, op requests, invite/key limit/unban requests/responses, and host additions. This setting lets you reduce the amount of logging. Valid settings: 2 to enable all logging, 1 to disable logging of ops/op requests, 0 to disable all logging except host additions."
  }

} ;# namespace botnetop

return
