###############################################################################
## This file is part of the Wolfpack module system
## Copyright (C) 1998-2001  Tothwolf <tothwolf@concentric.net>
## and is distributed according to the GNU general public license.
## For full details, read the top of wolfpack.tcl or the file called
## COPYING that was distributed with this code.
##
## $Id: usertools.tcl,v 1.6 2001/06/02 06:04:45 tothwolf Exp $
##

# name: usertools
# version: 1.0.0.0
# description: User and user data functions for use by other modules.
# provides: validHandle checkUser findUser findHandOnChannels findField \
#	clearFields wipeFields moveFields whoisFields getLastOn findChanFlag \
#	getBotName isBotHand chanFlag

namespace eval usertools {

  ##
  ## Check if a handle is valid and is not a bot
  ##
  ## Args:    handle
  ## Returns: 1 if the given handle is valid and is not a bot
  ##          0 otherwise
  ##
  proc validHandle {hand} {
    if {([validuser $hand]) && (![matchattr $hand b])} then {
      return 1
    }
    return 0
  }

  ##
  ## Sanity check the given handle
  ##
  ## Args:    handle, args {flags, channel}
  ## Returns: 0 if the given handle is not a valid user
  ##          -1 if the given handle is a bot
  ##          -2 if the given handle does not have the given flags
  ##          1 otherwise
  ##
  proc checkUser {hand args} {
    if {[validuser $hand]} then {
      if {[matchattr $hand b]} then {
        return -1
      } elseif {[string compare "" [set flags [lindex $args 0]]]} then {
        if {([string compare "" [set chan [lindex $args 1]]]) && \
            (![matchattr $hand $flags $chan])} then {
          return -2
        } elseif {![matchattr $hand $flags]} then {
          return -2
        }
      } else {
        return 1
      }
    }
    return 0
  }

  ##
  ## Find the handle of the given nick on a channel
  ##
  ## Args:    nick/handle, channel
  ## Returns: handle of the given nick
  ##
  proc findUser {who chan} {
    if {[validuser $who]} then {
      return [finduser [lindex [getuser $who HOSTS] 0]]
    } elseif {([validchan $chan]) && ([onchan $who $chan])} then {
      return [nick2hand $who $chan]
    } else {
      return
    }
  }

  ##
  ## Search for the given handle on all channels the bot monitors
  ##
  ## Args:    handle
  ## Returns: 1 if the given handle is on any channels the bot monitors
  ##          0 otherwise
  ##
  proc findHandOnChannels {hand} {
    foreach chan [channels] {
      if {[string compare "" [hand2nick $hand $chan]]} then {
        return 1
      }
    }
    return 0
  }

  ##
  ## Returns list of all users who have data in the given field
  ##
  ## Args:    field, data
  ## Returns: list of users
  ##
  proc findField {field {data ""}} {
    set users ""
    if {![string compare "" $data]} then {
      foreach user [lsort [userlist]] {
        if {[string compare "" [getuser $user XTRA $field]]} then {
          lappend users $user
        }
      }
    } else {
      foreach user [lsort [userlist]] {
        if {([string compare "" [set tmp [getuser $user XTRA $field]]]) && \
            (![string compare $tmp $data])} then {
          lappend users $user
        }
      }
    }
    return $users
  }

  ##
  ## Clear data for all the given fields for the given handle
  ##
  ## Args:    handle, fields
  ## Returns: nothing
  ##
  proc clearFields {hand fields} {
    foreach field $fields {
      if {[string compare "" [getuser $hand XTRA $field]]} then {
        setuser $hand XTRA $field
      }
    }
    return
  }

  ##
  ## Remove the given fields from all users
  ##
  ## Args:    fields
  ## Returns: nothing
  ##
  proc wipeFields {fields} {
    foreach user [userlist] {
      foreach field $fields {
        if {[string compare "" [getuser $user XTRA $field]]} then {
          setuser $user XTRA $field
        }
      }
    }
    return
  }

  ##
  ## Move data from one field to another
  ##
  ## Args:    old field, new field
  ## Returns: nothing
  ##
  proc moveFields {oldfield newfield} {
    foreach user [userlist] {
      if {[string compare "" [set data [getuser $user XTRA $oldfield]]]} then {
        setuser $user XTRA $newfield $data
        setuser $user XTRA $oldfield
      }
    }
    return
  }

  ##
  ## Add/remove/update fields in eggdrop's whois-fields variable
  ##
  ## Args:    action {add|remove|update}
  ## Returns: 1 if successful
  ##          0 otherwise
  ##
  proc whoisFields {action fields {oldFields ""}} {
    global whois-fields

    if {![info exists whois-fields]} then {
      set whois-fields ""
    }
    switch -exact -- $action {
      add -
      update {
        # Remove old fields from whois-fields
        if {[string compare "" $oldFields]} then {
          foreach oldField $oldFields {
            if {[lsearch -exact $fields $oldField] == -1} then {
              if {[set index [lsearch -exact ${whois-fields} $oldField]] != -1} then {
                set whois-fields [lreplace ${whois-fields} $index $index]
              }
            }
          }
        }
        # Add fields to whois-fields
        foreach field $fields {
          if {[lsearch -exact ${whois-fields} $field] == -1} then {
            lappend whois-fields $field
          }
        }
      }
      remove {
        # Remove fields from whois-fields
        foreach field $fields {
          if {[set index [lsearch -exact ${whois-fields} $field]] != -1} then {
            set whois-fields [lreplace ${whois-fields} $index $index]
          }
        }
      }
    }
    return
  }

  ##
  ## Get laston info for a user
  ##
  ## Args:    handle, args {channel}
  ## Returns: readable laston format for a user
  ##
  proc getLastOn {hand args} {
    set chan [lindex $args 0]
    if {![string compare "" $chan]} then {
      set laston [getuser $hand LASTON]
    } else {
      set laston [getuser $hand LASTON $chan]
    }
    set lastutime [lindex $laston 0]
    if {(![string compare "" $lastutime]) || \
        ($lastutime <= 0)} then {
      return "never nowhere"
    } else {
      if {[expr [clock seconds] - $lastutime] > 86400} then {
        set lasttime [clock format $lastutime -format "%d %b"]
      } else {
        set lasttime [clock format $lastutime -format "%H:%M"]
      }
      set lastchan [lindex $laston 1]
      if {![string compare "" $lastchan]} then {
        return [list $lasttime]
      } else {
        return [list $lasttime $lastchan]
      }
    }
  }

  ##
  ## Search all channel records for the given user for the given flag
  ##
  ## Args:   user, flag
  ## Returns: 1 if the given user has the given flag for any channel
  ##          0 otherwise
  ##
  proc findChanFlag {user flag} {
    foreach chan [channels] {
      if {[matchattr $user |$flag $chan]} then {
        return 1
      }
    }
    return 0
  }

  ##
  ## Find the bot's current nickname
  ##
  ## Args:    none
  ## Returns: bot's current nickname,
  ##          or desired nick if it dosen't currently have a nickname
  ##
  proc getBotName {} {
    global botnet-nick nick

    if {([info exists botnet-nick]) && \
        ([string compare "" ${botnet-nick}])} then {
      return ${botnet-nick}
    } else {
      return $nick
    }
  }

  ##
  ## Compare the given handle to the bot's handle
  ##
  ## Args:    hand
  ## Returns: 1 if hand matches bot's handle
  ##          0 otherwise
  ##
  proc isBotHand {hand} {
    if {![string compare [string toupper $hand] [string toupper [getBotName]]]} then {
      return 1
    }
    return 0
  }

  ##
  ## Find the most significant flag for the given user and channel
  ##
  ## Args:    user, channel
  ## Returns: most significant flag for the given user and channel
  ##
  proc chanFlag {user {chan ""}} {
    if {[validchan $chan]} then {
      set flags [chattr $user $chan]
    } else {
      set flags [chattr $user]
    }
    switch -glob -- $flags {
      "*b*\|*" {
        return b
      }
      "*n*\|*" {
        return N
      }
      "*\|*n*" {
        return n
      }
      "*m*\|*" {
        return M
      }
      "*\|*m*" {
        return m
      }
      "*d*\|*" {
        return D
      }
      "*\|*d*" {
        return d
      }
      "*a*\|*" {
        return A
      }
      "*\|*a*" {
        return a
      }
      "*o*\|*" {
        return O
      }
      "*\|*o*" {
        return o
      }
      "*q*\|*" {
        return Q
      }
      "*\|*q*" {
        return q
      }
      "*g*\|*" {
        return G
      }
      "*\|*g*" {
        return g
      }
      "*v*\|*" {
        return V
      }
      "*\|*v*" {
        return v
      }
    }
    return " "
  }

} ;# namespace usertools

return
